/* 
 * linux/arch/arm/drivers/sound/arcaudio.c
 *
 * Copyright (C) 1994, 1995, 1996 Russell King
 *
 * ARC audio driver
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>

#include <asm/arcaudio.h>
#include <asm/io.h>

#include "soundinit.h"
#include "soundcont.h"
#include "soundirq.h"
#include "voice.h"

#include <asm/segment.h>
#include <asm/hardware.h>

static int spkr;

static int arcaudio_write(struct inode *inode, struct file *filp, const char *buf, int nr)
{
	int retval=0,c;
	struct voice_driver *vd=(struct voice_driver *)filp->private_data;
	const char *b = buf;
	struct wait_queue wait;

	wait.task = current;
	wait.next = NULL;

	add_wait_queue(&vd->vd_wait, &wait);
	while(1) {
		current->state = TASK_INTERRUPTIBLE;
		if(current->signal & ~current->blocked) {
			retval = -ERESTARTSYS;
			break;
		}

		c = vd->vd_write(vd,b,nr);
		b += c;
		nr -= c;
		if(nr<=0)
			break;

		if (filp->f_flags & O_NONBLOCK) {
			retval = -EAGAIN;
			break;
		}
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&vd->vd_wait, &wait);
	return (b - buf) ? b - buf : retval;
}

static int arcaudio_open(struct inode *inode, struct file *filp)
{
	int voice;
	int channel;
	struct voice_driver *driver;
	
	channel = voice_findfree(0);
	
	if(channel < 0)
		return /* channel */-EBUSY;
	
	if((voice = voice_allocate(VOICE_SAMPLE, 0))<0)
		return voice;
	
	driver = (struct voice_driver *)voice;
	
	driver->vd_convtable = table_linsigned;
	driver->vd_bits      = 16;
	filp->private_data   = (void *)driver;
	voice_install(driver,channel);
	
	return 0;
}

static void arcaudio_release(struct inode *inode, struct file *filp)
{
	voice_release((struct voice_driver *)filp->private_data);
}

static int
arcaudio_getinfo(struct arcaudio *ainfo)
{
	int sr, nc;
	sound_configure(0,0,&sr,&nc);
	put_fs_long(sr, &ainfo->sample_rate);
	put_fs_long(nc, &ainfo->num_channels);
	put_fs_long(sound_vol(-1), &ainfo->volume);
	return 0;
}

static int
arcaudio_setinfo(struct arcaudio *ainfo)
{
	int ret;
  
	if((ret=sound_configure(get_fs_long(&ainfo->sample_rate), 
				get_fs_long(&ainfo->num_channels), NULL, NULL)) < 0)
		return ret;
	if((ret=sound_vol(get_fs_long(ainfo->volume))) < 0)
		return ret;
	return 0;
}

static int
arcaudio_getchannelinfo(struct voice_driver *vd, struct arcaudio_channel *ainfo)
{
	enum ch_type t;
	put_fs_long((sound_stereoimage[vd->vd_channel]-SOUND_CENTRE)<<24,
					&ainfo->stereo_position);
	put_fs_long(vd->vd_volume,&ainfo->channel_volume);
	switch(vd->vd_bits)
	{
		case 0: t = ARCAUDIO_NONE;          break;
		case 8: t = ARCAUDIO_8BITSIGNED;    break;
		case 9: t = ARCAUDIO_8BITUNSIGNED;  break;
		case 16:t = ARCAUDIO_16BITSIGNED;   break;
		case 17:t = ARCAUDIO_16BITUNSIGNED; break;
		default:t = ARCAUDIO_LOG;           break;
	}
	put_fs_long(t, &ainfo->channel_type);
	put_fs_long(vd->vd_size, &ainfo->buffer_size);
	return 0;
}

static int
arcaudio_setchannelinfo(struct voice_driver *vd, struct arcaudio_channel *ainfo)
{
	int newimage, newvol, newbits /* , newbuff */;
	
	newimage = ((get_fs_long(&ainfo->stereo_position)>>24) + SOUND_CENTRE)&255;
	if(newimage<0 || newimage>255)
		return -EINVAL;
	newvol = get_fs_long(&ainfo->channel_volume);
	if(newvol<0 || newvol>127)
		return -EINVAL;
	switch(get_fs_long(&ainfo->channel_type))
	{
		case ARCAUDIO_NONE:		newbits = 0;	break;
		case ARCAUDIO_8BITSIGNED:	newbits = 8;	break;
		case ARCAUDIO_8BITUNSIGNED:	newbits = 9;	break;
		case ARCAUDIO_16BITSIGNED:	newbits = 16;	break;
		case ARCAUDIO_16BITUNSIGNED:	newbits = 17;	break;
		case ARCAUDIO_LOG:		newbits = 1;	break;
		default:
			return -EINVAL;
	}
	
	sound_setimage(vd->vd_channel, newimage);
	vd->vd_bits = newbits;
	voice_volume(vd, newvol);
	return 0;
}

static int
arcaudio_getopts(unsigned long *arg)
{
	put_fs_long(spkr ? 1 : 0, arg);
	return 0;
}

static int
arcaudio_setopts(unsigned long *arg)
{
	spkr = get_fs_long(arg) & 1;
	if(spkr)
		outb((inb(IOC_CONTROL) & ~0x20) | 0xc0, IOC_CONTROL);
	else
		outb((inb(IOC_CONTROL) | 0x20) | 0xc0, IOC_CONTROL);
	return 0;
}

static int
arcaudio_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	switch(cmd)
	{
		case ARCAUDIO_GETINFO:
			return arcaudio_getinfo((struct arcaudio *)arg);
		
		case ARCAUDIO_SETINFO:
			return arcaudio_setinfo((struct arcaudio *)arg);
		
		case ARCAUDIO_GETCHANNELINFO:
			return arcaudio_getchannelinfo((struct voice_driver *)
					filp->private_data,(struct arcaudio_channel *)arg);
		
		case ARCAUDIO_SETCHANNELINFO:
			return arcaudio_setchannelinfo((struct voice_driver *)
					filp->private_data,(struct arcaudio_channel *)arg);

		case ARCAUDIO_GETOPTS:
			return arcaudio_getopts((unsigned long *)arg);
		case ARCAUDIO_SETOPTS:
			return arcaudio_setopts((unsigned long *)arg);
		default:
		  return -EINVAL;
	}
}

static struct file_operations arcaudio_fops =
{
/* 	arcaudio_lseek, */NULL,
/* 	arcaudio_read, */NULL,
	arcaudio_write,
	NULL,
/* 	arcaudio_select, */NULL,
	arcaudio_ioctl,
	NULL,
	arcaudio_open,
	arcaudio_release,
	NULL,
	NULL,
	NULL,
	NULL
};

void arcaudio_init(void)
{
  if(register_sounddev(16, "arcaudio", &arcaudio_fops, NULL))
    printk(KERN_ERR "Sound: can't register arc audio device\n");

  spkr = (inb(IOC_CONTROL) & 0x20) == 0;
}
