/*
 *  linux/arch/arm/kernel/setup.c
 *
 *  Copyright (C) 1995, 1996 Russell King
 */

/*
 * This file obtains various parameters about the system that the kernel
 * is running on.
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/ldt.h>
#include <linux/user.h>
#include <linux/a.out.h>
#include <linux/tty.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/major.h>
#include <linux/utsname.h>

#include <asm/segment.h>
#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/pgtable.h>
#include <asm/arch/mmu.h>
#include <asm/procinfo.h>

struct drive_info_struct { char dummy[32]; } drive_info;

struct screen_info screen_info = {
	0, 0,
	{ 0, 0 },
	0,
	0,
	100,
	0, 0, 0,
	75
};

struct param_struct {
    unsigned long page_size;
    unsigned long nr_pages;
    unsigned long ramdisk_size;
    unsigned long mountrootrdonly;
    unsigned long rootdev;
    unsigned long video_num_cols;
    unsigned long video_num_rows;
    unsigned long video_x;
    unsigned long video_y;
    unsigned long memc_control_reg;
    unsigned char sounddefault;
    unsigned char adfsdrives;
    unsigned char bytes_per_char_h;
    unsigned char bytes_per_char_v;
    unsigned long pages_in_bank[4];
    unsigned long unused[256/4-15];
    union {
	char paths[8][128];
	struct {
	    unsigned long magic;
	    char n[1024 - sizeof(unsigned long)];
	} s;
    } u;
    char commandline[256];
} *params = (struct param_struct *)PARAMS_BASE;

extern const struct processor_functions arm2_processor_functions;
extern const struct processor_functions arm250_processor_functions;
extern const struct processor_functions arm3_processor_functions;
extern const struct processor_functions arm6_processor_functions;
extern const struct processor_functions sa110_processor_functions;

struct armversions armidlist[] = {
#if defined(__arm2__) || defined(__arm3__)
	{ 0x41560200, 0xfffffff0, F_MEMC	, "ARM/VLSI",	"arm2"		, &arm2_processor_functions   },
	{ 0x41560250, 0xfffffff0, F_MEMC	, "ARM/VLSI",	"arm250"	, &arm250_processor_functions },
	{ 0x41560300, 0xfffffff0, F_MEMC|F_CACHE, "ARM/VLSI",	"arm3"		, &arm3_processor_functions   },
#endif
#if defined(__arm6__) || defined(__sa110__)
	{ 0x41560600, 0xffffff00, F_MMU|F_32BIT	, "ARM/VLSI",	"arm6"		, &arm6_processor_functions   },
	{ 0x41560610, 0xffffff00, F_MMU|F_32BIT	, "ARM/VLSI",	"arm610"	, &arm6_processor_functions   },
	{ 0x41007000, 0xffffff00, F_MMU|F_32BIT , "ARM/VLSI",   "arm7"		, &arm6_processor_functions   },
	{ 0x41007100, 0xffffff00, F_MMU|F_32BIT , "ARM/VLSI",   "arm710"	, &arm6_processor_functions   },
	{ 0x4401a100, 0xfffffff0, F_MMU|F_32BIT	, "DEC",	"sa110"		, &sa110_processor_functions  },
#endif
	{ 0x00000000, 0x00000000, 0		, "***",	"*unknown*"	, NULL }
};

#ifdef CONFIG_BLK_DEV_IMG
#if defined(CONFIG_BLK_DEV_IDE) || defined(CONFIG_BLK_DEV_HD)
#define DFLT_MAJOR IDE0_MAJOR
#elif defined(CONFIG_BLK_DEV_XD)
#define DFLT_MAJOR XT_DISK_MAJOR
#else
#define DFLT_MAJOR IDE0_MAJOR
#endif

static kdev_t image_dev[8] = {
    MKDEV(DFLT_MAJOR, 0x01),
    MKDEV(DFLT_MAJOR, 0x01),
    MKDEV(DFLT_MAJOR, 0x01),
    MKDEV(DFLT_MAJOR, 0x01),
    MKDEV(DFLT_MAJOR, 0x41),
    MKDEV(DFLT_MAJOR, 0x41),
    MKDEV(DFLT_MAJOR, 0x41),
    MKDEV(DFLT_MAJOR, 0x41)
};

extern void adfsimg_add_image (int slot, kdev_t dev, char *path);

static inline void img_setupdevs(struct param_struct *params)
{
	if (params->u.s.magic != 0xfeedface) {
		int i;
		for (i = 0; i < 8; i++)
			adfsimg_add_image(i, image_dev[i], params->u.paths[i]);
	} else {
		int i, p = 1;

		for (i = 0; i < params->u.s.n[0]; i++) {
			kdev_t dev;

			dev = MKDEV(params->u.s.n[p+1], params->u.s.n[p]);
			p += 2;
			if (MAJOR(dev) != UNNAMED_MAJOR)
				adfsimg_add_image(i, dev, &params->u.s.n[p]);
			p += strlen(&params->u.s.n[p]) + 1;
		}
	}
}
#else
#define img_setupdevs(x)
#endif

#ifdef CONFIG_ARCH_RPC
struct dram_bank drambank[NR_BANKS];
#endif

#define COMMAND_LINE_SIZE 256

unsigned long arm_id;
int armidindex;
int ioebpresent;
int memc_ctrl_reg;
int number_ide_drives;
int number_mfm_drives;

extern int bytes_per_char_h;
extern int bytes_per_char_v;
extern int root_mountflags;
extern int _etext, _edata, _end;
extern unsigned long real_end_mem;

#ifdef CONFIG_BLK_DEV_RAM
extern int rd_doload;		/* 1 = load ramdisk, 0 = don't load */
extern int rd_prompt;		/* 1 = prompt for ramdisk, 0 = don't prompt */
extern int rd_image_start;	/* starting block # of image */

static inline void setup_ramdisk (void)
{
    rd_image_start	= 0;
    rd_prompt		= 1;
    rd_doload		= 1;
}
#else
#define setup_ramdisk()	
#endif

static void check_ioeb_present(void)
{
    if (((*IOEB_BASE) & 15) == 5)
	armidlist[armidindex].features |= F_IOEB;
}

static char command_line[COMMAND_LINE_SIZE] = { 0, };
       char saved_command_line[COMMAND_LINE_SIZE];

struct processor_functions processor_functions;

void setup_arch(char **cmdline_p,
	unsigned long * memory_start_p, unsigned long * memory_end_p)
{
	unsigned long memory_start, memory_end;
	char c = ' ', *to = command_line, *from;
	int len = 0;

	armidindex = 0;
	while (1) {
		if (!((armidlist[armidindex].id ^ arm_id) &
			armidlist[armidindex].mask))
			break;
		armidindex += 1;
	}

	if (armidlist[armidindex].id == 0) {
		int i;
		for (i = 0; i < 3200; i++)
			((unsigned long *)SCREEN2_BASE)[i] = 0x77113322;
		while (1);
	}

	processor_functions = *armidlist[armidindex].procfns;

	check_ioeb_present();
	processor_functions._proc_init ();

	memory_start = MAPTOPHYS((unsigned long)&_end);
	memory_end = GET_MEMORY_END(params);

	bytes_per_char_h  = params->bytes_per_char_h;
	bytes_per_char_v  = params->bytes_per_char_v;
	from		  = params->commandline;
	ROOT_DEV	  = params->rootdev;
	ORIG_X		  = params->video_x;
	ORIG_Y		  = params->video_y;
	ORIG_VIDEO_COLS	  = params->video_num_cols;
	ORIG_VIDEO_LINES  = params->video_num_rows;
	memc_ctrl_reg	  = params->memc_control_reg;
	number_ide_drives = (params->adfsdrives >> 6) & 3;
	number_mfm_drives = (params->adfsdrives >> 3) & 3;

	setup_ramdisk();

#ifdef CONFIG_ARCH_RPC
	{
		int i;
		unsigned int bytes;

		for (i = 0, bytes = 0; i < NR_BANKS; i++) {
			drambank[i].v2p_offset = PAGE_OFFSET + bytes -
							(0x10000000 + (i << 26));
			bytes += params->pages_in_bank[i] * PAGE_SIZE;
			drambank[i].virt_offset = PAGE_OFFSET + bytes;
		}
		drambank[5].v2p_offset = 0xd6000000;
		drambank[5].virt_offset = 0xd8000000;
	}
#endif

	if (params->mountrootrdonly)
		root_mountflags |= MS_RDONLY;

	img_setupdevs(params);

	init_task.mm->start_code = TASK_SIZE;
	init_task.mm->end_code	 = TASK_SIZE + (unsigned long) &_etext;
	init_task.mm->end_data	 = TASK_SIZE + (unsigned long) &_edata;
	init_task.mm->brk	 = TASK_SIZE + (unsigned long) &_end;

    /* Save unparsed command line copy for /proc/cmdline */
    memcpy(saved_command_line, from, COMMAND_LINE_SIZE);
    saved_command_line[COMMAND_LINE_SIZE-1] = '\0';

    for (;;) {
	if (c == ' ' && *(unsigned long *)from == *(unsigned long *)"mem=") {
	    memory_end = simple_strtoul(from+4, &from, 0);
	    if ( *from == 'K' || *from == 'k' ) {
		memory_end = memory_end << 10;
		from++;
	    } else if ( *from == 'M' || *from == 'm' ) {
		memory_end = memory_end << 20;
		from++;
	    }
	    memory_end = memory_end + PAGE_OFFSET;
	}
	c = *(from++);
	if (!c)
	    break;
	if (COMMAND_LINE_SIZE <= ++len)
	    break;
	*(to++) = c;
    }

    *to = '\0';
    *cmdline_p = command_line;
    *memory_start_p = memory_start;
    *memory_end_p = memory_end;
    strcpy (system_utsname.machine, armidlist[armidindex].name);
}

#define ISSET(bit) (armidlist[armidindex].features & bit)

int get_cpuinfo(char * buffer)
{
	int len;

	len = sprintf (buffer,  "CPU:\n"
				"Type\t\t: %s\n"
				"Revision\t: %d\n"
				"Manufacturer\t: %s\n"
				"32bit modes\t: %s\n"
				"BogoMips\t: %lu.%02lu\n",
				armidlist[armidindex].name,
				(int)arm_id & 15,
				armidlist[armidindex].manu,
				ISSET (F_32BIT) ? "yes" : "no",
				(loops_per_sec+2500) / 500000,
				((loops_per_sec+2500) / 5000) % 100);
	len += sprintf (buffer + len,
				"\nHardware:\n"
				"Mem System\t: %s\n"
				"IOEB\t\t: %s\n",
				ISSET(F_MEMC)  ? "MEMC" : 
				ISSET(F_MMU)   ? "MMU"  : "*unknown*",
				ISSET(F_IOEB)  ? "present" : "absent"
				);
	return len;
}
